@@ -2,8 +2,9 @@ import re
 import copy
 import logging
 from pathlib import Path
-from dataclasses import dataclass
-from typing import List, Self, Dict, Any, Optional, Sequence
+from textwrap import dedent
+from dataclasses import dataclass, field
+from typing import List, Self, Dict, Any, Optional, Sequence, get_args
 
 import yaml
 
@@ -13,26 +14,26 @@ from jinja2 import Template
 from codearkt.python_executor import PythonExecutor
 from codearkt.tools import fetch_tools
 from codearkt.event_bus import AgentEventBus, EventType
-from codearkt.llm import LLM, ChatMessages, ChatMessage, FunctionCall, ToolCall
+from codearkt.llm import LLM, ChatMessages, ChatMessage
 from codearkt.util import get_unique_id
 
-
-END_CODE_SEQUENCE = "<end_code>"
-END_PLAN_SEQUENCE = "<end_plan>"
-STOP_SEQUENCES = [END_CODE_SEQUENCE, "Observation:", "Calling tools:"]
-DEFAULT_SERVER_URL = "http://localhost:5055"
-CURRENT_DIR = Path(__file__).parent
+DEFAULT_END_CODE_SEQUENCE = "<end_code>"
+DEFAULT_END_PLAN_SEQUENCE = "<end_plan>"
+DEFAULT_STOP_SEQUENCES = [DEFAULT_END_CODE_SEQUENCE, "Observation:", "Calling tools:"]
+AGENT_TOOL_PREFIX = "agent__"
 
 
 def extract_code_from_text(text: str) -> str | None:
-    pattern = r"[C|c]ode[\*]*:\n*```(?:py|python)?\s*\n(.*?)\n```"
+    pattern = r"[Cc]ode[\*]*\:[\*]*\s*\n*```(?:py|python)?\s*\n(.*?)\n```"
     matches = re.findall(pattern, text, re.DOTALL)
     if matches:
         return "\n\n".join(match.strip() for match in matches)
     return None
 
 
-def fix_code_actions(messages: List[ChatMessage]) -> List[ChatMessage]:
+def convert_code_to_content(
+    messages: List[ChatMessage], end_code_sequence: str
+) -> List[ChatMessage]:
     for message in messages:
         if message.role != "assistant":
             continue
@@ -48,7 +49,7 @@ def fix_code_actions(messages: List[ChatMessage]) -> List[ChatMessage]:
             if not code_action.endswith("```"):
                 code_action = code_action.rstrip("`")
                 code_action += "```"
-            code_action = code_action.strip() + END_CODE_SEQUENCE
+            code_action = code_action.strip() + end_code_sequence
             if isinstance(message.content, str):
                 message.content += "\n" + code_action
             else:
@@ -61,22 +62,38 @@ def fix_code_actions(messages: List[ChatMessage]) -> List[ChatMessage]:
 class Prompts:
     system: Template
     final: Template
+    no_code_action: Template
     plan: Optional[Template] = None
     plan_prefix: Optional[Template] = None
+    plan_suffix: Optional[Template] = None
+    end_code_sequence: str = DEFAULT_END_CODE_SEQUENCE
+    end_plan_sequence: str = DEFAULT_END_PLAN_SEQUENCE
+    stop_sequences: List[str] = field(default_factory=lambda: DEFAULT_STOP_SEQUENCES)
 
     @classmethod
     def load(cls, path: str | Path) -> Self:
         with open(path) as f:
             template = f.read()
         templates: Dict[str, Any] = yaml.safe_load(template)
-        wrapped_templates = {}
+        wrapped_templates: Dict[str, Any] = {}
         for key, value in templates.items():
-            wrapped_templates[key] = Template(value)
-        return cls(**wrapped_templates)
+            if value is None:
+                continue
+            field_type = cls.__annotations__.get(key)
+            if field_type is Template or Template in get_args(field_type):
+                wrapped_templates[key] = Template(value)
+            elif field_type is str:
+                wrapped_templates[key] = value.strip()
+            else:
+                wrapped_templates[key] = value
+        obj = cls(**wrapped_templates)
+        obj.stop_sequences = [s.strip() for s in obj.stop_sequences]
+        return obj
 
     @classmethod
     def default(cls) -> Self:
-        return cls.load(CURRENT_DIR / "prompts" / "default.yaml")
+        current_dir = Path(__file__).parent
+        return cls.load(current_dir / "prompts" / "default.yaml")
 
 
 class CodeActAgent:
@@ -90,7 +107,6 @@ class CodeActAgent:
         max_iterations: int = 10,
         verbosity_level: int = logging.ERROR,
         planning_interval: Optional[int] = None,
-        server_url: Optional[str] = DEFAULT_SERVER_URL,
         managed_agents: Optional[List[Self]] = None,
     ) -> None:
         self.name = name
@@ -101,11 +117,11 @@ class CodeActAgent:
         self.max_iterations = max_iterations
         self.verbosity_level = verbosity_level
         self.planning_interval = planning_interval
-        self.server_url = server_url
         self.managed_agents: Optional[List[Self]] = managed_agents
+
         if self.managed_agents:
             for agent in self.managed_agents:
-                agent_tool_name = "agent__" + agent.name
+                agent_tool_name = AGENT_TOOL_PREFIX + agent.name
                 if agent_tool_name not in self.tool_names:
                     self.tool_names.append(agent_tool_name)
 
@@ -128,33 +144,43 @@ class CodeActAgent:
         messages: ChatMessages,
         session_id: str,
         event_bus: AgentEventBus | None = None,
+        server_host: Optional[str] = None,
+        server_port: Optional[int] = None,
     ) -> str:
         messages = copy.deepcopy(messages)
 
-        # Register agent start
         run_id = get_unique_id()
         await self._publish_event(event_bus, session_id, EventType.AGENT_START)
         self._log(f"Starting agent {self.name}", run_id=run_id, session_id=session_id)
 
+        python_executor = None
         try:
-            # Initialize Python interpreter
             python_executor = PythonExecutor(
-                session_id=session_id, tool_names=self.tool_names, interpreter_id=run_id
+                session_id=session_id,
+                tool_names=self.tool_names,
+                interpreter_id=run_id,
+                tools_server_port=server_port,
+                tools_server_host=server_host,
             )
             self._log("Python interpreter started", run_id=run_id, session_id=session_id)
+            self._log(
+                f"Host: {server_host}, port: {server_port}",
+                run_id=run_id,
+                session_id=session_id,
+            )
 
-            # Check tools
-            tools = []
-            if self.server_url:
-                tools = await fetch_tools(self.server_url)
-                tools = [tool for tool in tools if tool.name in self.tool_names]
-                for tool_name in self.tool_names:
-                    assert tool_name in [tool.name for tool in tools], f"Tool {tool_name} not found"
+            tools = await self._get_tools(server_host=server_host, server_port=server_port)
+            self._log(
+                f"Fetched tools: {[tool.name for tool in tools]}",
+                run_id=run_id,
+                session_id=session_id,
+            )
             system_prompt = self.prompts.system.render(tools=tools)
-            self._log(f"Available tools: {self.tool_names}", run_id=run_id, session_id=session_id)
 
             # Form input messages
-            messages = fix_code_actions(messages)
+            messages = convert_code_to_content(
+                messages, end_code_sequence=self.prompts.end_code_sequence
+            )
             messages = [ChatMessage(role="system", content=system_prompt)] + messages
 
             for step_number in range(1, self.max_iterations + 1):
@@ -195,6 +221,7 @@ class CodeActAgent:
                 )
                 messages.extend(new_messages)
                 self._log("Final step completed", run_id=run_id, session_id=session_id)
+
         except Exception as e:
             self._log(
                 f"Agent {self.name} failed with error: {e}",
@@ -205,12 +232,35 @@ class CodeActAgent:
             raise e
         finally:
             # Cleanup
-            await python_executor.cleanup()
+            if python_executor:
+                await python_executor.cleanup()
             await self._publish_event(event_bus, session_id, EventType.AGENT_END)
 
         self._log(f"Agent {self.name} completed successfully", run_id=run_id, session_id=session_id)
         return str(messages[-1].content)
 
+    async def _get_tools(
+        self,
+        server_host: Optional[str],
+        server_port: Optional[int],
+    ) -> List[Tool]:
+        tools = []
+        fetched_tool_names = []
+        if server_host and server_port:
+            server_url = f"{server_host}:{server_port}"
+            tools = await fetch_tools(server_url)
+            for tool in tools:
+                if tool.description:
+                    tool.description = dedent(tool.description).strip()
+            tools = [tool for tool in tools if tool.name in self.tool_names]
+            fetched_tool_names = [tool.name for tool in tools]
+
+        for tool_name in self.tool_names:
+            assert (
+                tool_name in fetched_tool_names
+            ), f"Tool {tool_name} not found in {fetched_tool_names}"
+        return tools
+
     async def _step(
         self,
         messages: ChatMessages,
@@ -222,7 +272,7 @@ class CodeActAgent:
         self._log(
             f"Step inputs: {messages}", run_id=run_id, session_id=session_id, level=logging.DEBUG
         )
-        output_stream = self.llm.astream(messages, stop=STOP_SEQUENCES)
+        output_stream = self.llm.astream(messages, stop=self.prompts.stop_sequences)
 
         output_text = ""
         async for event in output_stream:
@@ -237,9 +287,9 @@ class CodeActAgent:
         if (
             output_text
             and output_text.strip().endswith("```")
-            and not output_text.strip().endswith(END_CODE_SEQUENCE)
+            and not output_text.strip().endswith(self.prompts.end_code_sequence)
         ):
-            chunk = END_CODE_SEQUENCE + "\n"
+            chunk = self.prompts.end_code_sequence + "\n"
             output_text += chunk
 
         self._log(
@@ -249,24 +299,30 @@ class CodeActAgent:
         new_messages = []
         if code_action is None:
             new_messages.append(ChatMessage(role="assistant", content=output_text))
+            if self._is_final_answer(output_text):
+                return new_messages
+            assert self.prompts.no_code_action is not None
+            no_code_action_prompt = self.prompts.no_code_action.render()
+            new_messages.append(ChatMessage(role="user", content=no_code_action_prompt))
+            self._log(
+                no_code_action_prompt,
+                run_id=run_id,
+                session_id=session_id,
+                level=logging.DEBUG,
+            )
             return new_messages
 
         self._log(
             f"Code action: {code_action}", run_id=run_id, session_id=session_id, level=logging.DEBUG
         )
-        tool_call_id = f"toolu_{get_unique_id()}"
         tool_call_message = ChatMessage(
             role="assistant",
             content=output_text,
-            tool_calls=[
-                ToolCall(
-                    id=tool_call_id,
-                    function=FunctionCall(name="python_interpreter", arguments=code_action),
-                )
-            ],
         )
         await self._publish_event(event_bus, session_id, EventType.TOOL_CALL, code_action)
         new_messages.append(tool_call_message)
+
+        # Execute code
         try:
             code_result = await python_executor.invoke(code_action)
             self._log(
@@ -275,20 +331,25 @@ class CodeActAgent:
                 session_id=session_id,
                 level=logging.DEBUG,
             )
-            code_result_messages = code_result.to_messages(tool_call_id)
-            new_messages.extend(code_result_messages)
-            tool_output: str = str(code_result_messages[0].content) + "\n"
+            code_result_message: ChatMessage = code_result.to_message()
+            assert isinstance(code_result_message.content, list)
+            new_messages.append(code_result_message)
+            tool_output: str = str(code_result_message.content[0]["text"]) + "\n"
             await self._publish_event(event_bus, session_id, EventType.TOOL_RESPONSE, tool_output)
         except Exception as e:
-            new_messages.append(
-                ChatMessage(role="tool", content=f"Error: {e}", tool_call_id=tool_call_id)
-            )
+            new_messages.append(ChatMessage(role="user", content=f"Error: {e}"))
             self._log(f"Code error: {e}", run_id=run_id, session_id=session_id, level=logging.DEBUG)
             await self._publish_event(
                 event_bus, session_id, EventType.TOOL_RESPONSE, f"Error: {e}\n"
             )
         return new_messages
 
+    def _is_final_answer(self, content: str) -> bool:
+        code_action = extract_code_from_text(content)
+        assert code_action is None
+        content = content.lower()
+        return "final" in content and "answer" in content
+
     async def _handle_final_message(
         self,
         messages: ChatMessages,
@@ -306,7 +367,7 @@ class CodeActAgent:
             level=logging.DEBUG,
         )
 
-        output_stream = self.llm.astream(input_messages, stop=STOP_SEQUENCES)
+        output_stream = self.llm.astream(input_messages, stop=self.prompts.stop_sequences)
         output_text = ""
         async for event in output_stream:
             if isinstance(event.content, str):
@@ -315,6 +376,7 @@ class CodeActAgent:
                 chunk = "\n".join([str(item) for item in event.content])
             output_text += chunk
             await self._publish_event(event_bus, session_id, EventType.OUTPUT, chunk)
+
         self._log(
             f"Final message: {final_message}",
             run_id=run_id,
@@ -335,12 +397,15 @@ class CodeActAgent:
         assert (
             self.prompts.plan_prefix is not None
         ), "Plan prefix is not set, but planning is enabled"
+        assert (
+            self.prompts.plan_suffix is not None
+        ), "Plan suffix is not set, but planning is enabled"
 
-        conversation = "\n\n".join([f"{m.role}: {m.content}" for m in messages])
+        conversation = "\n\n".join([f"{m.role}: {m.content}" for m in messages[1:]])
         planning_prompt = self.prompts.plan.render(conversation=conversation, tools=tools)
         input_messages = [ChatMessage(role="user", content=planning_prompt)]
 
-        output_stream = self.llm.astream(input_messages, stop=[END_PLAN_SEQUENCE])
+        output_stream = self.llm.astream(input_messages, stop=[self.prompts.end_plan_sequence])
 
         plan_prefix = self.prompts.plan_prefix.render().strip() + "\n\n"
         await self._publish_event(event_bus, session_id, EventType.OUTPUT, plan_prefix)
@@ -351,7 +416,10 @@ class CodeActAgent:
             chunk = event.content
             output_text += chunk
             await self._publish_event(event_bus, session_id, EventType.OUTPUT, chunk)
-        await self._publish_event(event_bus, session_id, EventType.OUTPUT, "\n\n")
+
+        plan_suffix = "\n\n" + self.prompts.plan_suffix.render().strip() + "\n\n"
+        await self._publish_event(event_bus, session_id, EventType.OUTPUT, plan_suffix)
+        output_text += plan_suffix
 
         return [ChatMessage(role="assistant", content=output_text)]
 
