@@ -6,18 +6,20 @@ import atexit
 import signal
 import json
 from typing import Optional, Any, List, Dict, Sequence
+import threading
 
 import docker
 from docker.models.containers import Container
 from docker.client import DockerClient
+from docker.models.networks import Network
 from pydantic import BaseModel
 
 from codearkt.llm import ChatMessage
 from codearkt.tools import fetch_tools
-from codearkt.util import get_unique_id
+from codearkt.util import get_unique_id, truncate_content, is_correct_json
 
 
-IMAGE: str = "phoenix120/codearkt_http"
+IMAGE: str = "phoenix120/codearkt_http:v2"
 MEM_LIMIT: str = "512m"
 CPU_QUOTA: int = 50000
 CPU_PERIOD: int = 100000
@@ -29,16 +31,18 @@ CONTAINER_NAME: str = "codearkt_http"
 
 _CLIENT: Optional[DockerClient] = None
 _CONTAINER: Optional[Container] = None
+_DOCKER_LOCK: threading.Lock = threading.Lock()
 
 
 def cleanup_container(signum: Optional[Any] = None, frame: Optional[Any] = None) -> None:
     global _CONTAINER
-    if _CONTAINER:
-        try:
-            _CONTAINER.remove(force=True)
-            _CONTAINER = None
-        except Exception:
-            pass
+    with _DOCKER_LOCK:
+        if _CONTAINER:
+            try:
+                _CONTAINER.remove(force=True)
+                _CONTAINER = None
+            except Exception:
+                pass
     if signum == signal.SIGINT:
         raise KeyboardInterrupt()
 
@@ -48,34 +52,12 @@ signal.signal(signal.SIGINT, cleanup_container)
 signal.signal(signal.SIGTERM, cleanup_container)
 
 
-MAX_LENGTH_TRUNCATE_CONTENT: int = 20000
-
-
-def truncate_content(content: str, max_length: int = MAX_LENGTH_TRUNCATE_CONTENT) -> str:
-    if len(content) <= max_length:
-        return content
-    else:
-        return (
-            content[: max_length // 2]
-            + f"\n..._This content has been truncated to stay below {max_length} characters_...\n"
-            + content[-max_length // 2 :]
-        )
-
-
-def is_correct_json(content: str) -> bool:
-    try:
-        json.loads(content)
-        return True
-    except json.JSONDecodeError:
-        return False
-
-
 class ExecResult(BaseModel):  # type: ignore
     stdout: str
     error: str | None = None
     result: Any | None = None
 
-    def to_messages(self, tool_call_id: str) -> List[ChatMessage]:
+    def to_message(self) -> ChatMessage:
         image_content: List[Dict[str, Any]] | None = None
         output: str = "Stdout:\n" + self.stdout + "\n\n"
         if self.result is not None:
@@ -86,7 +68,7 @@ class ExecResult(BaseModel):  # type: ignore
                     image_content = [
                         {
                             "type": "image_url",
-                            "image_url": {"url": f"data:image/jpeg;base64,{image_base64}"},
+                            "image_url": {"url": f"data:image/png;base64,{image_base64}"},
                         }
                     ]
             except json.JSONDecodeError:
@@ -97,11 +79,12 @@ class ExecResult(BaseModel):  # type: ignore
             output += "Error: " + self.error
         output = output.strip()
 
-        messages = []
-        messages.append(ChatMessage(role="tool", content=output, tool_call_id=tool_call_id))
+        content = []
+        if output:
+            content.append({"type": "text", "text": output})
         if image_content:
-            messages.append(ChatMessage(role="user", content=image_content))
-        return messages
+            content += image_content
+        return ChatMessage(role="user", content=content)
 
 
 def init_docker() -> docker.DockerClient:
@@ -127,86 +110,112 @@ def init_docker() -> docker.DockerClient:
     return client
 
 
+def run_network(client: DockerClient) -> Network:
+    try:
+        net = client.networks.get(NET_NAME)
+    except docker.errors.NotFound:
+        net = client.networks.create(
+            NET_NAME,
+            driver="bridge",
+            internal=False,
+            options={
+                "com.docker.network.bridge.enable_ip_masquerade": "false",
+            },
+        )
+    return net
+
+
+def run_container(client: DockerClient, net_name: str) -> Container:
+    return client.containers.run(
+        IMAGE,
+        name=CONTAINER_NAME,
+        detach=True,
+        auto_remove=True,
+        ports={"8000/tcp": None},
+        mem_limit=MEM_LIMIT,
+        cpu_period=CPU_PERIOD,
+        cpu_quota=CPU_QUOTA,
+        pids_limit=PIDS_LIMIT,
+        cap_drop=["ALL"],
+        read_only=True,
+        tmpfs={"/tmp": "rw,size=64m", "/run": "rw,size=16m"},
+        security_opt=["no-new-privileges"],
+        extra_hosts={"host.docker.internal": "host-gateway"},
+        sysctls={"net.ipv4.ip_forward": "0"},
+        network=net_name,
+        dns=[],
+    )
+
+
 class PythonExecutor:
     def __init__(
         self,
         tool_names: Sequence[str] = tuple(),
         session_id: Optional[str] = None,
-        mcp_server_port: int = 5055,
+        tools_server_host: Optional[str] = None,
+        tools_server_port: Optional[int] = None,
         interpreter_id: Optional[str] = None,
     ) -> None:
         global _CLIENT, _CONTAINER
 
-        if not _CLIENT:
-            _CLIENT = init_docker()
-        client = _CLIENT
+        with _DOCKER_LOCK:
+            if not _CLIENT:
+                _CLIENT = init_docker()
+            client = _CLIENT
 
-        if _CONTAINER is None:
-            try:
-                _CONTAINER = _CLIENT.containers.get(CONTAINER_NAME)
-            except docker.errors.NotFound:
+            if not _CONTAINER:
                 try:
-                    net = client.networks.get(NET_NAME)
+                    _CONTAINER = client.containers.get(CONTAINER_NAME)
                 except docker.errors.NotFound:
-                    net = client.networks.create(
-                        NET_NAME,
-                        driver="bridge",
-                        internal=False,
-                        options={
-                            "com.docker.network.bridge.enable_ip_masquerade": "false",
-                        },
-                    )
+                    net = run_network(client)
+                    _CONTAINER = run_container(client, str(net.name))
 
-                _CONTAINER = client.containers.run(
-                    IMAGE,
-                    name=CONTAINER_NAME,
-                    detach=True,
-                    auto_remove=True,
-                    ports={"8000/tcp": None},
-                    mem_limit=MEM_LIMIT,
-                    cpu_period=CPU_PERIOD,
-                    cpu_quota=CPU_QUOTA,
-                    pids_limit=PIDS_LIMIT,
-                    cap_drop=["ALL"],
-                    read_only=True,
-                    tmpfs={"/tmp": "rw,size=64m", "/run": "rw,size=16m"},
-                    security_opt=["no-new-privileges"],
-                    extra_hosts={"host.docker.internal": "host-gateway"},
-                    sysctls={"net.ipv4.ip_forward": "0"},
-                    network=net.name,
-                    dns=[],
-                    environment={"SERVER_PORT": str(mcp_server_port)},
-                )
-
-        self.container = _CONTAINER
-        self.mcp_server_port = mcp_server_port
+            self.container = _CONTAINER
+        self.tools_server_host = tools_server_host
+        self.tools_server_port = tools_server_port
         self.session_id = session_id
         self.interpreter_id: str = interpreter_id or get_unique_id()
         self.tool_names = tool_names
         self.tools_are_checked = False
         self.url = self._get_url()
+        self.is_ready = False
 
     async def invoke(self, code: str) -> ExecResult:
-        await self._check_tools()
-        await self._wait_for_ready()
+        if not self.tools_are_checked:
+            await self._check_tools()
+            self.tools_are_checked = True
+
+        if not self.is_ready:
+            await self._wait_for_ready()
+            self.is_ready = True
+
         result = await self._call_exec(code)
         return result
 
-    async def _check_tools(self) -> None:
-        if not self.tool_names or self.tools_are_checked:
-            return
-        available_tools = await fetch_tools(f"http://localhost:{self.mcp_server_port}")
-        available_tool_names = [tool.name for tool in available_tools]
-        for tool_name in self.tool_names:
-            if tool_name not in available_tool_names:
-                raise ValueError(f"Tool {tool_name} not found in MCP server")
-        self.tools_are_checked = True
+    def _are_tools_available(self) -> bool:
+        return bool(self.tool_names and self.tools_server_host and self.tools_server_port)
 
-    async def _call_exec(self, code: str) -> ExecResult:
+    async def _check_tools(self) -> None:
+        assert not self.tools_are_checked
+
+        available_tool_names = []
+        if self._are_tools_available():
+            server_url = f"{self.tools_server_host}:{self.tools_server_port}"
+            available_tools = await fetch_tools(server_url)
+            available_tool_names = [tool.name for tool in available_tools]
+
+        for tool_name in self.tool_names:
+            if tool_name.startswith("agent__"):
+                continue
+            if tool_name not in available_tool_names:
+                raise ValueError(f"Tool {tool_name} not found in {available_tool_names}")
+
+    async def _call_exec(self, code: str, send_tools: bool = True) -> ExecResult:
         payload = {
             "code": textwrap.dedent(code),
             "session_id": self.session_id,
-            "tool_names": self.tool_names,
+            "tool_server_port": self.tools_server_port,
+            "tool_names": self.tool_names if send_tools and self._are_tools_available() else [],
             "interpreter_id": self.interpreter_id,
         }
 
@@ -248,7 +257,7 @@ class PythonExecutor:
         start_time = time.time()
         while time.time() - start_time < max_wait:
             try:
-                output = await self._call_exec("print('ready')")
+                output = await self._call_exec("print('ready')", send_tools=False)
                 assert output.stdout.strip() == "ready"
                 return
             except (httpx.RequestError, httpx.TimeoutException, AssertionError):
